; -------------------------------------
; 	Z80 Virtual Machine v4
;   (c) 2015 - 2017 kio@little-bat.de
; -------------------------------------


; Registers:
;
; AF = temp
; BC = VM PC
; DE = VM vstack Top value
; HL = temp
; IX = VM rstack pointer
; SP = VM vstack pointer
; IY = Address of next
; AF' .. HL' = not used


#include "core-macros.s"


#local


#assert BC_IS_VPC	    == 1
#assert SP_IS_VSTACK	    == 1
#assert IX_IS_RSTACK	    == 1
#assert DE_IS_TOP	    == 1
;#assert HEAP_RETAIN_RELEASE == 1
;#assert HEAP_USES_HANDLES == 1
;#assert IY_IS_NEXT == 1



; ________________________________________________________
; enter VM coded procedure:
; push current VM program counter (BC) on VM return stack (IX) 
; and switch to VM mode:

p_enter::
	pushr_bc
	;jr	virtual
	
; ________________________________________________________
; switch from Z80 native to VM mode:
; pop Z80 IP from Z80 rstack into VM IP and jump to next opcode:

virtual::
	pop	bc
	next


#if IY_IS_NEXT
#code GSINIT		; reusable label 'next' must be used after definition:
	ld  iy, next	; use register IY for 'jp (iy)' instead of 'jp next'
#code VMCORE
#endif



; -----------------------------------------
; Code Flow:
; -----------------------------------------

; ________________________________
; switch to Z80 native code:
; switch back to virtual mode with macro VIRTUAL
; during native mode register BC can be used
;
; NATIVE ( -- )

NATIVE::	ld	hl,bc
		jp	hl

; _______________________________
; return from virtual procedure:
;
; RETURN ( -- )
; DROPxRETURN ( -- )
; NIPxRETURN ( -- )

DROP8RETURN::	pop	de
DROP7RETURN::	pop	de
DROP6RETURN::	pop	de
DROP5RETURN::	pop	de
DROP4RETURN::	pop	de
DROP3RETURN::	pop	de
DROP2RETURN::	pop	de
DROP1RETURN::	pop	de
DROP0RETURN::
RETURN::	popr_bc
NOP::		next

NIP8RETURN::	pop	af
NIP7RETURN::	pop	af
NIP6RETURN::	pop	af
NIP5RETURN::	pop	af
NIP4RETURN::	pop	af
NIP3RETURN::	pop	af
NIP2RETURN::	pop	af
NIP1RETURN::	pop	af
NIP0RETURN::	popr_bc
		next

DROPxRETURN:: 	ld	a,(bc++)	;13	a = (x-1)*2
		ld	l,a		;4
		ld	h,0		;7
		add	hl,sp		;10
		ld	sp,hl		;4
		jr	DROP1RETURN	;22	= 60cc

DROPxRETURN0::	ld	de,0		; 	a = (x-1)*2
NIPxRETURN::	ld	a,(bc++)	;13	a = x*2
		ld	l,a		;4
		ld	h,0		;7
		add	hl,sp		;10
		ld	sp,hl		;4
		jr	RETURN		;12	= 50cc



; _______________________________
; call proc ptr
;
; CALLPROCPTR ( [args] addr -- [rval] )

CALLPROCPTR::	ex	hl,de
		pop	de
		jp	hl

; ________________________________
; JP opcodes

; JP     ( $addr.w -- )	    ; jp to label
; JP0    ( n $addr.w -- )   ; jp to label if n==0
; JP1    ( n $addr.w -- )   ; jp to label if n!=0
; AND0JP ( n $addr.w -- )   ; keep n and jp to label if n==0 else drop n & don't jp
; OR1JP  ( n $addr.w -- )   ; keep n and jp to label if n!=0 else drop n & don't jp


JP0::		ld	a,d
		or	e
		pop	de
		jr	z,JP

skip2::		inc	bc
		inc	bc
		next

JP1::		ld	a,d
		or	e
		pop	de
		jr	z,skip2
		;jr	JP

JP::		ld	a,(bc++)
		ld	l,a
		ld	a,(bc++)
		ld	b,a
		ld	c,l
		next

AND0JP::	ld	a,d
		or	e
		jr	z,JP
		pop	de
		jr	skip2

OR1JP::		ld	a,d
		or	e
		jr	nz,JP
		pop	de
		jr	skip2

; ________________________________
; BRA opcodes with negative offset:
; jr with offset = dis.b - $100
; the offset is relative to the position of the offset byte

; JRM  ( $dis.b -- )
; JRM0 ( n $dis.b -- )		; jr if n=FALSE
; JRM1 ( n $dis.b -- )		; jr if n=TRUE

JR0m::		ld	a,d
		or	e
		pop	de
		jr	z,JRm
		jr	skip1

JR1m::		ld	a,d
		or	e
		pop	de
		jr	z,skip1
		;jr	JRm

JRm::		ld	a,(bc)
		add	a,c
		ld	c,a
		jr	c,1$
		dec	b
1$:		next

; ________________________________
; BRA opcodes with positive offset:
; the offset is relative to the position of the offset byte

; JRP  ( $dis.b -- )
; JRP0 ( n $dis.b -- )	; jr if n==0
; JRP1 ( n $dis.b -- )	; jr if n!=0
; AND0 ( n $dis.b -- )	; keep n and jr if n==0 else drop n & don't jp
; OR1  ( n $dis.b -- )	; keep n and jr if n!=0 else drop n & don't jp

JR0p::		ld	a,d
		or	e
		pop	de
		jr	z,JRp
		;jr	skip1
skip1:		inc	bc
		next

JR1p::		ld	a,d
		or	e
		pop	de
		jr	z,skip1
		;jr	JRp

JRp::		ld	a,(bc)
		add	a,c
		ld	c,a
		jr	nc,1$
		inc	b
1$:		next

AND0::		ld	a,d
		or	e
		jr	z,JRp
		pop	de
		jr	skip1

OR1::		ld	a,d
		or	e
		jr	nz,JRp
		pop	de
		jr	skip1

; ________________________________
; if n<L[].count jump to label L[n] else jump to default label
; jump table can have from 1 to 256 entries for case 0 to 255
;
; SWITCH ( n $tablesize.b $defaultlabel.w $table.w[] -- )
;
; invocation:
;
;	dw	SWITCH
;	db	tablesize-1
;	dw	default_label
;	dw	table[tablesize]; list of labels

SWITCH::	ld	hl,bc		; hl -> tablesize.b
		ld	a,(hl++)	; a = tablesize-1, hl -> default_label

		rlc	d		; test de.hi
		jr	nz,1$		; de.hi != 0 -> oorange

		cp	a,e		; sz-1 - idx
		jr	c,1$		; sz-1<idx -> sz≤idx -> idx≥sz -> beyond table

		; de inside table
		ex	hl,de		; hl=idx
		inc	hl		; +1 to skip over default entry
		add	hl,hl		; hl=idx*2 for word-size entries
		add	hl,de		; add table base

1$:		ld	c,(hl++)	; get entry: low byte ..
		ld	b,(hl)		; .. high byte

		pop_de_next		; pop new TOP and jump to label


; -----------------------------------------
; 	MOVE TO / FROM RSTACK
; -----------------------------------------

; TOR     ( n16 -- )		; move word from vstack to rstack
; TOR32   ( n32 -- )		; move int32 or two int16 to rstack
; DUP2R   ( n16 -- n16 )	; copy top value from vstack to rstack
; DUP2R32 ( n32 -- n32 )	; copy top long value to rstack
; FROMR   ( -- n16 )		; move word from rstack to vstack
; FROMR32 ( -- n32 )		; move int32 from rstack to vstack
; PEEKR   ( -- n16 )		; copy top value from rstack to vstack

TOR32::		pushr_de
		pop	de
TOR::		pushr_de
		pop	de
		jp_next

DUP2R::		pushr_de
		jr_next

DUP2R32::	pop	hl
		push	hl
		pushr_de
		pushr_hl
		jr_next

FROMR32::	push	de
		popr_de
FROMR::		push	de
		popr_de
		jr_next

PEEKR::		push	de
		peekr_de
		jr_next


; -----------------------------------------
; 	PUSH IMMEDIATE VALUE
; -----------------------------------------

; IVAL    ( $.w -- n16 )	; push word value
; IVALm1  ( -- n16 )		; push immediate value -1
; IVAL0   ( -- n16 )		; push immediate value 0
; IVAL1   ( -- n16 )		; push immediate value 1
; IVAL2   ( -- n16 )		; push immediate value 2
; IVAL3   ( -- n16 )		; push immediate value 3
; IVALu8  ( $.b -- n16 )	; push positive word value (hi byte := 0x00)
; IVALs8  ( $.b -- n16 )	; push negative word value (hi byte := 0xff)
; IVAL32  ( $.l -- n32 )	; push long value: attn.: high word first!
; IVALu32 ( $.w -- n32 )	; push positive long value (hi word := 0x0000)
; IVALs32 ( $.w -- n32 )	; push negative long value (hi word := 0xffff)

IVALs32::	ld	a,0xff
		db	opcode(cp a,N)	; jr u32

IVALu32::	xor	a
u32:		push	de
		ld	e,a
		;ld	d,a
		jr	i32

IVAL32::	push	de
		ld	a,(bc++)
		ld	e,a
		ld	a,(bc++)
i32:		ld	d,a		; d = n.hi
		;jr	IVAL

IVAL::		push	de
		ld	a,(bc++)
		ld	e,a
		ld	a,(bc++)
		ld	d,a
		next

IVALm1::	push	de
return_m1::	ld	de,-1		; -1
		jr_next

IVAL0l::	push	de
		ld	de,0
		;jr	IVAL0

IVAL0::		push	de
return_0::	ld	de,0
		next

IVAL1l::	push	de
		ld	de,0
		;jr	IVAL1

IVAL1::		push	de
return_1::	ld	de,1
		next

IVAL2::		push	de
		ld	de,2
		jr_next

IVAL3::		push	de
		ld	de,3
		jr_next

IVALs8::	push	de
		ld	a,(bc++)
		ld	e,a
extend_neg::	ld	d,0xff
		jr_next

IVALu8::	push	de
		ld	a,(bc++)
;ld_de_a_next:: deprecated
return_a::	ld	e,a
LOu16u8::
CASTu16u8::
extend_pos::	ld	d,0	;7	=45cc
		next		;38
HIu16u8::	ld	e,d
		jr	extend_pos

; boolean helpers:
drop2_ret_nz::	pop	de
drop1_ret_nz::	pop	de
return_nz::	jr	nz,return_1	; optimized for case nz wg. int32 compares
		jr	return_0

drop2_ret_z::	pop	de
drop1_ret_z::	pop	de
return_z::	jr	nz,return_0	; optimized for case nz wg. int32 compares
		jr	return_1

drop2_ret_hl_pos::add	hl,hl
drop2_ret_nc::	pop	de
drop1_ret_nc::	pop	de
return_nc::	jr	c,return_0
		jr	return_1

drop2_ret_hl_neg::add	hl,hl
drop2_ret_cy::	pop	de
drop1_ret_cy::	pop	de
return_cy::	jr	nc,return_0
		jr	return_1


; -----------------------------------------
; 	STACK BASICS
; -----------------------------------------

; DUP   ( n16 -- n16 n16 )		; LGET 0;   duplicate top value
; DUP32 ( n32 -- n32 n32 )		; LGET32 0; duplicate top long value
; OVER  ( n2 n1 -- n2 n1 n2 )  		; LGET 2
; SWAP  ( n1 n2 -- n2 n1 ) 		; swap two words on stack

SWAP::		pop	hl
		jr	push_de_exhlde_next

DUP32::		pop	hl
		push	hl
		push	de
		push	hl
		;jr	next
		db	opcode(cp a,N)

push_de_next::
DUP::		push	de		; 49 cc
		next

OVER::		pop	hl		; 74 cc
		push	hl
push_de_exhlde_next::
		push	de
exhlde_next::	ex	hl,de
		next



; DROP	     ( n16 -- )		; drop top value
; DROP2      ( n32 -- )		; drop 2 words
; DROP3      ( n n n -- )	; drop 3 words
; DROP4      ( n n n n -- )	; drop 4 words
; DROPx      ( … %n.b -- )	; drop n-1 words
; NIP 	     ( b a -- a )	; drop 2nd word on stack
; NIP2       ( c b a -- a )	; drop 2 words behind Top
; NIP3       ( d c b a -- a )	; drop 3 words behind Top
; NIP4       ( e d c b a -- a )	; drop 4 words behind Top
; NIPx       ( … a %n.b -- a )	; drop n words behind Top

;DROPx:: 	ld	a,(bc++)	; 13	a = (x-1)*2
;		ld	l,a		; 4
;		ld	h,0		; 7
;		add	hl,sp		; 10
;		ld	sp,hl		; 4	= 38 cc -> DROPx lohnt sich erst ab DROP6
;		jr	DROP1
;
;NIPx::		ld	a,(bc++)	; 13	a = x*2
;		ld	l,a		; 4
;		ld	h,0		; 7
;		add	hl,sp		; 10
;		ld	sp,hl		; 4	= 38 cc -> DROPx lohnt sich erst ab NIP5
;		jr	NIP0

DROP8::		pop	de
DROP7::		pop	de
DROP6::		pop	de
DROP5::		pop	de
DROP4::		pop	de
DROP3::		pop	de
DROP2::		pop	de
pop_de_next:: = $
DROP1::
DROP::		pop	de
DROP0::		next

NIP8::		pop	af
NIP7::		pop	af
NIP6::		pop	af
NIP5::		pop	af
NIP4::		pop	af
NIP3::		pop	af
NIP2::		pop	af
NIP1::
NIP::		pop	af
NIP0::		next


#endlocal


